/* func.c   by Michael Thorpe   2005-07-11 */

#include <string.h>
#include "lisp.h"

/* This function does not incref/decref anything; returns 0 on success */
static int bindlist(scope *s,obj *names,obj *values) {
	obj *tmp;

	if(names==&nilobj) {
		if(values==&nilobj)
			return(0);
		return(-1);
	}
	if(names->type==&cons_objtype) {
		if(values->type != &cons_objtype)
			return(-1);
		if(bindlist(s,CAR(names),CAR(values)))
			return(-1);
		return(bindlist(s,CDR(names),CDR(values)));
	}
	tmp=define_local(incscoperef(s),incref(names),incref(values));
	if(tmp) {
		decref(tmp);
		return(0);
	}
	return(-1);
}

static obj *applyfunc(scope *s,obj *func,obj *args) {
	obj *prog;
	scope *t;

	decscoperef(s);
	t=newscope(incscoperef(globalscope));
	if(!t) {
		decref(func);
		decref(args);
		return(0);
	}
	if(bindlist(t,func->value.func.args,args)) {
		decscoperef(t);
		decref(args);
		return(throwargumenterror(func));
	}
	prog=incref(func->value.func.prog);
	decref(func);
	decref(args);
	return(evaluate(t,prog));
}

static obj *func_tostring(obj *in) {
	obj *args,*func;
	obj *out=0;

	args=(in->value.func.args->type->tostring)(incref(in->value.func.args));
	if(args) {
		func=(in->value.func.prog->type->tostring)(incref(in->value.func.prog));
		if(func) {
			out=newstring(0,8+args->value.s.len+1+func->value.s.len+1);
			if(out) {
				memcpy(out->value.s.data,"(lambda ",8);
				memcpy(out->value.s.data+8,args->value.s.data,args->value.s.len);
				out->value.s.data[8+args->value.s.len]=' ';
				memcpy(out->value.s.data+8+args->value.s.len+1,func->value.s.data,func->value.s.len);
				out->value.s.data[8+args->value.s.len+1+func->value.s.len]=')';
			}
			decref(func);
		}
		decref(args);
	}
	decref(in);
	return(out);
}

static void drop_func(obj *o) {
	decref(o->value.func.args);
	decref(o->value.func.prog);
}

objtype func_objtype={
	name: "func",
	tostring: func_tostring,
	dropobj: drop_func,
	applyfunc: applyfunc
};

