(#define '#f (= 1 2))
(#define '#t (= 1 1))
(#define 'pair? (#lambda '(o) '(eqv? 'cons (object-type o))))
(#define 'stream? (#lambda '(o) '(eqv? 'stream (object-type o))))
(#define 'car (#lambda '((a . d)) 'a))
(#define 'cdr (#lambda '((a . d)) 'd))
(#define 'caar (#lambda '(((aa . da) . rest)) 'aa))
(#define 'cadr (#lambda '((a ad . dd)) 'ad))
(#define 'cdar (#lambda '(((aa . da) . rest)) 'da))
(#define 'cddr (#lambda '((a ad . dd)) 'dd))
(#define 'list (#lambda 'aaa 'aaa))
(#define 'cons (#lambda '(a b) '(list a . b)))

(#define 'get
	(#lambda '(hash item)
		'(#ifelse hash
			'(#ifelse (eqv? (caar hash) item)
				'(cdar hash)
				'(get (cdr hash) item))
			'#f)))

(#define 'macroexpansions '())
(#define 'dont-expand '(dont expand))
; Should this be rewritten to handle tail recursion?
(#define 'macro-expand
	(#lambda '(o)
		'(#ifelse (pair? o)
			'(#prog2of2
				(#prog2of2
					(#define-local 'macro (get macroexpansions (car o)))
					'(#ifelse macro
						'(#prog2of2
							(#prog2of2
								(#define-local 'cdr-o (cdr o))
								'(#define-local 'expansion (macro . cdr-o)))
							'(#ifelse (eq? expansion dont-expand)
								'#f
								'(#define-local 'o (macro-expand expansion))))
						'#f))
				'(#map '(macro-expand #object) o))
			'o)))
(#define '#macro-define
	(#lambda '(macro func)
		'(#define 'macroexpansions (cons (cons macro func) macroexpansions))))

(#macro-define 'define
	(#lambda '(name . value)
		'(#ifelse (pair? name)
			'(list '#define (quote (car name)) (list 'lambda (cdr name) (cons 'progn value)))
			'(list '#define (quote name) . value))))
(#macro-define 'define-local
	(#lambda '(name . value)
		'(#ifelse (pair? name)
			'(list '#define-local (quote (car name)) (list 'lambda (cdr name) (cons 'progn value)))
			'(list '#define-local (quote name) . value))))
(#macro-define 'lambda
	(#lambda '(args . prog)
		'(list '#lambda (quote args) (quote (macro-expand (cons 'progn prog))))))
(#macro-define 'macro-define
	(#lambda '(name . value)
		'(#ifelse (pair? name)
			'(list '#macro-define (quote (car name)) (list 'lambda (cdr name) (cons 'progn value)))
			'(list '#macro-define (quote name) . value))))
(#macro-define 'progn
	(#lambda 'aaa
		'(#ifelse aaa
			'(#ifelse (pair? aaa)
				'(#ifelse (pair? (cdr aaa))
					'(list '#prog2of2 (car aaa) (quote (macro-expand (cons 'progn (cdr aaa)))))
					'(car aaa))
				'aaa)
			'aaa)))

(#macro-define 'catch
	(#lambda '(cond body handler)
		'(list '#catch cond
			(quote (macro-expand body))
			(quote (macro-expand handler)))))
(#macro-define 'ifelse
	(#lambda '(cond iftrue iffalse)
		'(list '#ifelse cond (quote (macro-expand iftrue)) (quote (macro-expand iffalse)))))
(#macro-define 'loop
	(#lambda 'body
		'(list '#loop (quote (macro-expand (cons 'progn body))))))
(#macro-define 'map
	(#lambda '(mapper . args)
		'(list '#map (quote (macro-expand body)) . args)))
(#macro-define 'subscope
	(#lambda 'body
		'(list '#subscope (quote (macro-expand (cons 'progn body))))))

(#define 'evaluate-file
	(#lambda '(file)
		'(#catch 'end-of-file
			'(#loop '(evaluate (macro-expand (read-object file))))
			'#f)))

(evaluate-file #preload.lisp)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; From this point on we're operating on our own repl (minus the printing) ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(macro-define (and a . b)
	(ifelse b
		(list 'ifelse a (cons 'and b) #f)
		a))
(macro-define (exit-loop) (list 'throw ''exit-loop #f))
(macro-define (if bool . progs)
	(list 'ifelse bool
		(cons 'progn progs)
		#f))
(macro-define (next-loop) (list 'throw ''next-loop #f))
(macro-define (not a) (list 'ifelse a #f #t))
(macro-define (or a . b)
	(ifelse b
		(list 'subscope
			(list 'define-local 'a a)
			(list 'ifelse 'a 'a (cons 'or b)))
		a))
(macro-define (unless bool . progs)
	(list 'ifelse bool
		#f
		(cons 'progn progs)))
(macro-define (while cond . body)
	(list 'loop
		(list 'ifelse cond
			(cons 'progn body)
			(list 'throw ''exit-loop #f))))

(define (dprint o)
	(write-object #standard-out o)
	(write-char #standard-out 10))

(define (write-newline file)
	(write-char file 10))
(define (show-error error errorvalue)
	(write-newline #standard-out)
	(write-string #standard-out |EXCEPTION: |)
	(write-object #standard-out error)
	(write-newline #standard-out)
	(write-string #standard-out |EXCEPTION-VALUE: |)
	(write-object #standard-out errorvalue)
	(write-newline #standard-out))
(define (read-eval-print in out prompt)
	(loop
		(if prompt (write-string out prompt))
		(define-local scratchspace
			(catch 'end-of-file
				(read-object in)
				(progn (if prompt (write-newline out)) (exit-loop))))
		(define-local scratchspace
			(catch #f
				(evaluate (macro-expand scratchspace))
				(progn (show-error #exception #exceptionvalue) (ifelse prompt (next-loop) (exit-loop)))))
		(if prompt
			(write-object out scratchspace)
			(write-newline out))))
(read-eval-print #standard-in #standard-out (ifelse (isatty #standard-in) |mikelisp> | #f))
